// slice set of simplices

#include <fstream>
#include <iostream>
#include <string>

#include "ijkIO.txx"

typedef float COORD_TYPE;

typedef enum OUTPUT_FORMAT { OFF, IV };

// global variables
int dimension = 3;
int numv = 0;
float * vertex_coord = NULL;
int nums = 0;
int * simplex_vert = NULL;
int nume = 0;
int slice_axis = 0;
float slice_coord;
int num_intersected_edges = 0;
int * intersected_edge_vert = NULL;
int num_slice_simplices = 0;
int * slice_simplex_vert = NULL;
const char * isotable_directory = NULL;
OUTPUT_FORMAT output_format = OFF;
bool standard_input = false;
char * input_filename = NULL;
char * output_filename = NULL;

void parse_command_line(int argc, char **argv);
void usage_error();

namespace IJKSLICE {
void slice_simplices
(const int vertex_dimension, 
 const COORD_TYPE * vertex_coord, const int num_vert, 
 const int simplex_dimension,
 const int * simplex_vert, const int num_simplices, 
 const int slice_axis, const COORD_TYPE slice_coord,
 const char * isotable_directory,
 int * & intersected_edge_vert, int & num_intersected_edges,
 int * & slice_simplex_vert, int & num_slice_simplices);
}

using namespace std;

int main(int argc, char **argv)
{
  isotable_directory = getenv("IJK_ISOTABLE_DIR");

  #ifdef IJK_ISOTABLE_DIR
  if (isotable_directory == NULL) {
    isotable_directory = IJK_ISOTABLE_DIR;
  }
  #endif

  parse_command_line(argc, argv);

  if (isotable_directory == NULL) {
    cerr << "Error.  Unknown isotable directory." << endl;
    cerr << "  Use -dir {isotable_directory} argument or set environment variable IJK_ISOTABLE_DIR." << endl;
    exit(5);
  };

  try {
    IJKIO::ERROR error_slice("ijkslice::main");

    istream * in;
    if (input_filename == NULL)
      in = &cin;
    else {
      in = new ifstream(input_filename, ios::in);
      if (!in->good()) {
	cerr << "Error. Unable to read file " << input_filename << "." 
	     << endl;
	cerr << "Exiting." << endl;
	exit(12);
      }
    };

    IJKIO::ijkinOFF(*in, dimension, vertex_coord, numv, simplex_vert, nums);

    if (slice_axis >= dimension)
      throw error_slice("Slice axis must be less than dimension.");
  }
  catch(IJKIO::ERROR error) {
    cerr << "Error in procedure: " << error.procedure << endl;
    cerr << "  " << error.msg1 << endl;
    exit(15);
  };

  IJKSLICE::slice_simplices
    (dimension, vertex_coord, numv, dimension-1, simplex_vert, nums,
     slice_axis, slice_coord, isotable_directory,
     intersected_edge_vert, num_intersected_edges,
     slice_simplex_vert, num_slice_simplices);

  int num_slice_vert = num_intersected_edges;
  COORD_TYPE * coord = new COORD_TYPE[(dimension-1)*num_slice_vert];
  int * coord0 = new int[dimension-1];
  int * coord1 = new int[dimension-1];

  for (int iv = 0; iv < num_slice_vert; iv++) {
    int v0 = intersected_edge_vert[2*iv];
    int v1 = intersected_edge_vert[2*iv+1];
    float c0 = vertex_coord[v0*dimension+slice_axis];
    float c1 = vertex_coord[v1*dimension+slice_axis];
    float s0 = (c1 - slice_coord)/(c1-c0);
    float s1 = (c0 - slice_coord)/(c0-c1);
    int d2 = 0;
    for (int d = 0; d < dimension; d++) {
      if (d != slice_axis) {
	float c = s0*vertex_coord[v0*dimension+d]+
	  s1*vertex_coord[v1*dimension+d];
	coord[iv*(dimension-1) + d2] = c;
	d2++;
      };
    };
  }
  delete [] coord0;
  delete [] coord1;

  try {

    IJKIO::ERROR error_slice("ijkslice");

    ostream * out = NULL;
    if (output_filename == NULL)
      out = & cout;
    else
      out = new ofstream(output_filename, ios::out);

    switch (output_format) {

    case OFF:
      IJKIO::ijkoutOFF(*out, dimension-1, coord, num_slice_vert, 
		       slice_simplex_vert, num_slice_simplices);
      break;

    case IV:
      if (dimension == 4) {
	IJKIO::ijkoutIV(*out, dimension-1, coord, num_slice_vert, 
			slice_simplex_vert, num_slice_simplices);
      }
      else throw error_slice("Illegal dimension. OpenInventor format is only for dimension 3.");
      break;

    default:
      throw error_slice("Illegal output format.");
      break;
    }
  }
  catch(IJKIO::ERROR error) {
    cerr << "Error in procedure: " << error.procedure << endl;
    cerr << "  " << error.msg1 << endl;
  }

  delete [] intersected_edge_vert;
  delete [] simplex_vert;
  delete [] coord;
  delete [] vertex_coord;
}

void parse_command_line(int argc, char **argv)
{
  int iarg = 1;
  while (iarg < argc && argv[iarg][0] == '-') {
    if (strcmp(argv[iarg], "-axis") == 0) {
      iarg++;
      if (iarg >= argc) usage_error();
      sscanf(argv[iarg], "%d", &slice_axis);
    }
    else if (strcmp(argv[iarg], "-x") == 0) { slice_axis = 0; }
    else if (strcmp(argv[iarg], "-y") == 0) { slice_axis = 1; }
    else if (strcmp(argv[iarg], "-z") == 0) { slice_axis = 2; }
    else if (strcmp(argv[iarg], "-w") == 0) { slice_axis = 3; }
    else if (strcmp(argv[iarg], "-cin") == 0) {
      standard_input = true;
    }
    else if (strcmp(argv[iarg], "-dir") == 0) {
      iarg++;
      if (iarg >= argc) usage_error();
      isotable_directory = argv[iarg];
    }
    else
      usage_error();

    iarg++;
  };

  if ((argc - iarg) < 1) usage_error();
  sscanf(argv[iarg], "%f", &slice_coord);
  iarg++;

  if ((argc - iarg) > 0) { 
    input_filename = argv[iarg]; 
    iarg++;
  };

  if ((argc - iarg) > 0) { 
    output_filename = argv[iarg]; 
    iarg++;
  };

  if ((argc - iarg) != 0) { usage_error(); }

  if (!standard_input && input_filename == NULL) {
    cerr << "Error: Use -cin flag to read from standard input or provide an input file." << endl;
    usage_error();
  };
}

void usage_error()
{
  cerr << "Usage: ijkslice [-axis {n}|-x|-y|-z|-w] [-cin] {slice x-coord} [input file] [output file]" << endl;
  exit(10);
}
